@@ -35,20 +35,17 @@
 #include <QGraphicsScene>
 #include <QGraphicsPixmapItem>
 #include <QGraphicsDropShadowEffect>
+#include <QSet>
 
 #include <KWindowSystem>
 #include <KWindowInfo>
-#include <KScreen/Config>
-#include <KScreen/GetConfigOperation>
-#include <KScreen/Output>
 
 #include <xcb/xcb_cursor.h>
 #include <xcb/xcb_util.h>
 #include <xcb/xfixes.h>
 
 X11ImageGrabber::X11ImageGrabber(QObject *parent) :
-    ImageGrabber(parent),
-    mScreenConfigOperation(nullptr)
+    ImageGrabber(parent)
 {
     mNativeEventFilter = new OnClickEventFilter(this);
 }
@@ -85,7 +82,6 @@ bool OnClickEventFilter::nativeEventFilter(const QByteArray &eventType, void *me
 
             {
                 xcb_button_release_event_t *ev2 = static_cast<xcb_button_release_event_t *>(message);
-                qDebug() << ev2->detail;
                 if (ev2->detail == 1) {
                     QMetaObject::invokeMethod(mImageGrabber, "doImageGrab", Qt::QueuedConnection);
                 } else if (ev2->detail < 4) {
@@ -195,6 +191,14 @@ QPixmap X11ImageGrabber::convertFromNative(xcb_image_t *xcbImage)
         return QPixmap(); // we don't know
     }
 
+    // The RGB32 format requires data format 0xffRRGGBB, ensure that this fourth byte really is 0xff
+    if (format == QImage::Format_RGB32) {
+        quint32 *data = reinterpret_cast<quint32 *>(xcbImage->data);
+        for (int i = 0; i < xcbImage->width * xcbImage->height; i++) {
+            data[i] |= 0xff000000;
+        }
+    }
+
     QImage image(xcbImage->data, xcbImage->width, xcbImage->height, format);
 
     if (image.isNull()) {
@@ -209,24 +213,20 @@ QPixmap X11ImageGrabber::convertFromNative(xcb_image_t *xcbImage)
         image.setColor(1, QColor(Qt::black).rgb());
     }
 
-    // done
-
-    return QPixmap::fromImage(image);
+    // Image is ready. Since the backing data from xcbImage could be freed
+    // before the QPixmap goes away, a deep copy is necessary.
+    return QPixmap::fromImage(image).copy();
 }
 
 // utility functions
 
+// Note: x, y, width and height are measured in device pixels
 QPixmap X11ImageGrabber::blendCursorImage(const QPixmap &pixmap, int x, int y, int width, int height)
 {
-    // first we get the cursor position, compute the co-ordinates of the region
-    // of the screen we're grabbing, and see if the cursor is actually visible in
-    // the region
+    // If the cursor position lies outside the area, do not bother drawing a cursor.
 
-    const qreal dpr = pixmap.devicePixelRatio();
-
-    // cursor position operates on application's device pixel ratio, not the pixmap!
-    QPoint cursorPos = QCursor::pos() / dpr * qApp->devicePixelRatio();
-    QRect screenRect(x / dpr, y / dpr, width / dpr, height / dpr);
+    QPoint cursorPos = getNativeCursorPosition();
+    QRect screenRect(x, y, width, height);
 
     if (!screenRect.contains(cursorPos)) {
         return pixmap;
@@ -250,15 +250,14 @@ QPixmap X11ImageGrabber::blendCursorImage(const QPixmap &pixmap, int x, int y, i
     // process the image into a QImage
 
     QImage cursorImage = QImage((quint8 *)pixelData, cursorReply->width, cursorReply->height, QImage::Format_ARGB32_Premultiplied);
-    cursorImage.setDevicePixelRatio(dpr);
 
     // a small fix for the cursor position for fancier cursors
 
-    cursorPos -= QPoint(cursorReply->xhot, cursorReply->yhot) / dpr;
+    cursorPos -= QPoint(cursorReply->xhot, cursorReply->yhot);
 
     // now we translate the cursor point to our screen rectangle
 
-    cursorPos -= QPoint(x, y) / dpr;
+    cursorPos -= QPoint(x, y);
 
     // and do the painting
 
@@ -271,46 +270,75 @@ QPixmap X11ImageGrabber::blendCursorImage(const QPixmap &pixmap, int x, int y, i
     return blendedPixmap;
 }
 
-QPixmap X11ImageGrabber::getWindowPixmap(xcb_window_t window, bool blendPointer)
+QPixmap X11ImageGrabber::getPixmapFromDrawable(xcb_drawable_t drawableId, const QRect &rect)
 {
     xcb_connection_t *xcbConn = QX11Info::connection();
 
-    // first get geometry information for our drawable
+    // proceed to get an image based on the geometry (in device pixels)
 
-    xcb_get_geometry_cookie_t geomCookie = xcb_get_geometry_unchecked(xcbConn, window);
-    CScopedPointer<xcb_get_geometry_reply_t> geomReply(xcb_get_geometry_reply(xcbConn, geomCookie, NULL));
-
-    // then proceed to get an image
-
-    CScopedPointer<xcb_image_t> xcbImage(
+    QScopedPointer<xcb_image_t, ScopedPointerXcbImageDeleter> xcbImage(
         xcb_image_get(
             xcbConn,
-            window,
-            geomReply->x,
-            geomReply->y,
-            geomReply->width,
-            geomReply->height,
+            drawableId,
+            rect.x(),
+            rect.y(),
+            rect.width(),
+            rect.height(),
             ~0,
             XCB_IMAGE_FORMAT_Z_PIXMAP
         )
     );
 
-    // if the image is null, this means we need to get the root image window
-    // and run a crop
-
+    // too bad, the capture failed.
     if (xcbImage.isNull()) {
-        return getWindowPixmap(QX11Info::appRootWindow(), blendPointer)
-                .copy(geomReply->x, geomReply->y, geomReply->width, geomReply->height);
+        return QPixmap();
     }
 
     // now process the image
 
     QPixmap nativePixmap = convertFromNative(xcbImage.data());
+    return nativePixmap;
+}
+
+// finalize the grabbed pixmap where we know the absolute position
+QPixmap X11ImageGrabber::postProcessPixmap(QPixmap pixmap, QRect rect, bool blendPointer)
+{
     if (!(blendPointer)) {
-        return nativePixmap;
+        // note: this may be the null pixmap if an error occurred.
+        return pixmap;
     }
 
-    // now we blend in a pointer image
+    return blendCursorImage(pixmap, rect.x(), rect.y(), rect.width(), rect.height());
+}
+
+QPixmap X11ImageGrabber::getToplevelPixmap(QRect rect, bool blendPointer)
+{
+    xcb_window_t rootWindow = QX11Info::appRootWindow();
+
+    // Treat a null rect as an alias for capturing fullscreen
+    if (!rect.isValid()) {
+        rect = getDrawableGeometry(rootWindow);
+    }
+
+    QPixmap nativePixmap = getPixmapFromDrawable(rootWindow, rect);
+    return postProcessPixmap(nativePixmap, rect, blendPointer);
+}
+
+QPixmap X11ImageGrabber::getWindowPixmap(xcb_window_t window, bool blendPointer)
+{
+    xcb_connection_t *xcbConn = QX11Info::connection();
+
+    // first get geometry information for our window
+
+    xcb_get_geometry_cookie_t geomCookie = xcb_get_geometry_unchecked(xcbConn, window);
+    CScopedPointer<xcb_get_geometry_reply_t> geomReply(xcb_get_geometry_reply(xcbConn, geomCookie, NULL));
+    QRect rect(geomReply->x, geomReply->y, geomReply->width, geomReply->height);
+
+    // then proceed to get an image
+
+    QPixmap nativePixmap = getPixmapFromDrawable(window, rect);
+
+    // Translate window coordinates to global ones.
 
     xcb_get_geometry_cookie_t geomRootCookie = xcb_get_geometry_unchecked(xcbConn, geomReply->root);
     CScopedPointer<xcb_get_geometry_reply_t> geomRootReply(xcb_get_geometry_reply(xcbConn, geomRootCookie, NULL));
@@ -320,7 +348,16 @@ QPixmap X11ImageGrabber::getWindowPixmap(xcb_window_t window, bool blendPointer)
     CScopedPointer<xcb_translate_coordinates_reply_t> translateReply(
         xcb_translate_coordinates_reply(xcbConn, translateCookie, NULL));
 
-    return blendCursorImage(nativePixmap, translateReply->dst_x,translateReply->dst_y, geomReply->width, geomReply->height);
+    // Adjust local to global coordinates.
+    rect.moveRight(rect.x() + translateReply->dst_x);
+    rect.moveTop(rect.y() + translateReply->dst_y);
+
+    // If the window capture failed, try to obtain one from the full screen.
+    if (nativePixmap.isNull()) {
+        return getToplevelPixmap(rect, blendPointer);
+    }
+
+    return postProcessPixmap(nativePixmap, rect, blendPointer);
 }
 
 bool X11ImageGrabber::isKWinAvailable()
@@ -335,57 +372,20 @@ bool X11ImageGrabber::isKWinAvailable()
     return false;
 }
 
-void X11ImageGrabber::KWinDBusScreenshotHelper(quint64 window)
+void X11ImageGrabber::KWinDBusScreenshotHelper(quint64 pixmapId)
 {
-    mPixmap = getWindowPixmap((xcb_window_t)window, false);
-    emit pixmapChanged(mPixmap);
-}
-
-void X11ImageGrabber::KScreenCurrentMonitorScreenshotHelper(KScreen::ConfigOperation *op)
-{
-    KScreen::ConfigPtr config = qobject_cast<KScreen::GetConfigOperation *>(op)->config();
-
-    if (!config)           { return grabFullScreen(); }
-    if (!config->screen()) { return grabFullScreen(); }
-
-    // we'll store the cursor position first
-
-    QPoint cursorPosition = QCursor::pos();
-
-    // next, we'll get all our outputs and figure out which one has the cursor
-
-    const KScreen::OutputList outputs = config->outputs();
-    for (auto output: outputs) {
-        if (!(output->isConnected())) { continue; }
-        if (!(output->currentMode())) { continue; }
-
-        QPoint screenPosition = output->pos();
-        QSize  screenSize     = output->currentMode()->size();
-        QRect  screenRect     = QRect(screenPosition, screenSize);
-
-        if (!(screenRect.contains(cursorPosition))) {
-            continue;
-        }
-
-        // bingo, we've found an output that contains the cursor. Now
-        // to take a shot
-
-        mPixmap = getWindowPixmap(QX11Info::appRootWindow(), mCapturePointer);
-        mPixmap = mPixmap.copy(screenPosition.x(), screenPosition.y(), screenSize.width(), screenSize.height());
+    // obtain width and height and grab an image (x and y are always zero for pixmaps)
+    QRect rect = getDrawableGeometry((xcb_drawable_t)pixmapId);
+    mPixmap = getPixmapFromDrawable((xcb_drawable_t)pixmapId, rect);
+    if (!mPixmap.isNull()) {
         emit pixmapChanged(mPixmap);
-
-        mScreenConfigOperation->disconnect();
-        mScreenConfigOperation->deleteLater();
-        mScreenConfigOperation = nullptr;
-
         return;
     }
 
-    mScreenConfigOperation->disconnect();
-    mScreenConfigOperation->deleteLater();
-    mScreenConfigOperation = nullptr;
-
-    return grabFullScreen();
+    // Cannot retrieve pixmap from KWin, just fallback to fullscreen capture. We
+    // could try to detect the original action (window under cursor or active
+    // window), but that is too complex for this edge case.
+    grabFullScreen();
 }
 
 void X11ImageGrabber::rectangleSelectionCancelled()
@@ -415,7 +415,7 @@ void X11ImageGrabber::rectangleSelectionConfirmed(const QPixmap &pixmap, const Q
 
 void X11ImageGrabber::grabFullScreen()
 {
-    mPixmap = getWindowPixmap(QX11Info::appRootWindow(), mCapturePointer);
+    mPixmap = getToplevelPixmap(QRect(), mCapturePointer);
     emit pixmapChanged(mPixmap);
 }
 
@@ -423,51 +423,43 @@ void X11ImageGrabber::grabTransientWithParent()
 {
     xcb_window_t curWin = getRealWindowUnderCursor();
 
-    // do we have a top-level or a transient window?
-
-    KWindowInfo info(curWin, NET::WMName, NET::WM2TransientFor | NET::WM2WindowClass);
-    if (!(info.valid(true) && (info.transientFor() != XCB_WINDOW_NONE)) ||
-          info.windowClassClass().isEmpty() || info.windowClassName().isEmpty()) {
-        return grabWindowUnderCursor();
-    }
-
     // grab the image early
 
-    mPixmap = getWindowPixmap(QX11Info::appRootWindow(), false);
+    mPixmap = getToplevelPixmap(QRect(), false);
 
     // now that we know we have a transient window, let's
-    // see if the parent has any other transient windows who's
-    // transient for the same app
-
+    // find other possible transient windows and the app window itself.
     QRegion clipRegion;
-    QStack<xcb_window_t> childrenStack = findAllChildren(findParent(curWin));
 
-    while (!(childrenStack.isEmpty())) {
-        xcb_window_t winId = childrenStack.pop();
-        KWindowInfo tempInfo(winId, 0, NET::WM2TransientFor);
+    QSet<xcb_window_t> transients;
+    xcb_window_t parentWinId = curWin;
+    do {
+        // find parent window and add the window to the visible region
+        xcb_window_t winId = parentWinId;
+        QRect winRect;
+        parentWinId = getTransientWindowParent(winId, winRect);
+        transients << winId;
+        clipRegion += winRect;
 
-        if (info.transientFor() == tempInfo.transientFor()) {
-            clipRegion += getApplicationWindowGeometry(winId);
-        }
-    }
+        // Continue walking only if this is a transient window (having a parent)
+    } while (parentWinId != XCB_WINDOW_NONE && !transients.contains(parentWinId));
 
-    // now we have a list of all the transient windows for the
-    // parent, time to find the parent
+
+    // All parents are known now, find other transient children.
+    // Assume that the lowest window is behind everything else, then if a new
+    // transient window is discovered, its children can then also be found.
 
     QList<WId> winList = KWindowSystem::stackingOrder();
-    for (int i = winList.size() - 1; i >= 0; i--) {
-        KWindowInfo tempInfo(winList[i], NET::WMGeometry | NET::WMFrameExtents, NET::WM2WindowClass);
+    for (auto winId : winList) {
+        QRect winRect;
+        xcb_window_t parentWinId = getTransientWindowParent(winId, winRect);
 
-        QString winClass(tempInfo.windowClassClass());
-        QString winName(tempInfo.windowClassName());
-
-        if (winClass.contains(info.name(), Qt::CaseInsensitive) || winName.contains(info.name(), Qt::CaseInsensitive)) {
-            if (mCaptureDecorations) {
-                clipRegion += tempInfo.frameGeometry();
-            } else {
-                clipRegion += tempInfo.geometry();
+        // if the parent should be displayed, then show the child too
+        if (transients.contains(parentWinId)) {
+            if (!transients.contains(winId)) {
+                transients << winId;
+                clipRegion += winRect;
             }
-            break;
         }
     }
 
@@ -593,7 +585,7 @@ void X11ImageGrabber::grabApplicationWindowHelper(xcb_window_t window)
     KWindowInfo info(window, NET::WMFrameExtents);
     if (info.valid()) {
         QRect frameGeom = info.frameGeometry();
-        mPixmap = getWindowPixmap(QX11Info::appRootWindow(), mCapturePointer).copy(frameGeom);
+        mPixmap = getToplevelPixmap(frameGeom, mCapturePointer);
     }
 
     // fallback is window without the frame
@@ -601,11 +593,11 @@ void X11ImageGrabber::grabApplicationWindowHelper(xcb_window_t window)
     emit pixmapChanged(mPixmap);
 }
 
-QRect X11ImageGrabber::getApplicationWindowGeometry(xcb_window_t window)
+QRect X11ImageGrabber::getDrawableGeometry(xcb_drawable_t drawable)
 {
     xcb_connection_t *xcbConn = QX11Info::connection();
 
-    xcb_get_geometry_cookie_t geomCookie = xcb_get_geometry_unchecked(xcbConn, window);
+    xcb_get_geometry_cookie_t geomCookie = xcb_get_geometry_unchecked(xcbConn, drawable);
     CScopedPointer<xcb_get_geometry_reply_t> geomReply(xcb_get_geometry_reply(xcbConn, geomCookie, NULL));
 
     return QRect(geomReply->x, geomReply->y, geomReply->width, geomReply->height);
@@ -613,14 +605,27 @@ QRect X11ImageGrabber::getApplicationWindowGeometry(xcb_window_t window)
 
 void X11ImageGrabber::grabCurrentScreen()
 {
-    mScreenConfigOperation = new KScreen::GetConfigOperation;
-    connect(mScreenConfigOperation, &KScreen::GetConfigOperation::finished,
-            this, &X11ImageGrabber::KScreenCurrentMonitorScreenshotHelper);
+    QPoint cursorPosition = QCursor::pos();
+    for (auto screen : QGuiApplication::screens()) {
+        QRect screenRect = screen->geometry();
+        if (!screenRect.contains(cursorPosition)) {
+            continue;
+        }
+
+        // The screen origin is in native pixels, but the size is device-dependent. Convert these also to native pixels.
+        QRect nativeScreenRect(screenRect.topLeft(), screenRect.size() * screen->devicePixelRatio());
+        mPixmap = getToplevelPixmap(nativeScreenRect, mCapturePointer);
+        emit pixmapChanged(mPixmap);
+        return;
+    }
+
+    // No screen found with our cursor, fallback to capturing full screen
+    grabFullScreen();
 }
 
 void X11ImageGrabber::grabRectangularRegion()
 {
-    QuickEditor *editor = new QuickEditor(getWindowPixmap(QX11Info::appRootWindow(), false));
+    QuickEditor *editor = new QuickEditor(getToplevelPixmap(QRect(), false));
 
     connect(editor, &QuickEditor::grabDone, this, &X11ImageGrabber::rectangleSelectionConfirmed);
     connect(editor, &QuickEditor::grabCancelled, this, &X11ImageGrabber::rectangleSelectionCancelled);
@@ -677,29 +682,31 @@ xcb_window_t X11ImageGrabber::getRealWindowUnderCursor()
     return pointerReply->child;
 }
 
-QStack<xcb_window_t> X11ImageGrabber::findAllChildren(xcb_window_t window)
+
+// obtain the size of the given window, returning the window ID of the parent
+xcb_window_t X11ImageGrabber::getTransientWindowParent(xcb_window_t winId, QRect &outRect)
 {
-    QStack<xcb_window_t> winStack;
-    xcb_connection_t *xcbConn = QX11Info::connection();
+    NET::Properties properties = mCaptureDecorations ? NET::WMFrameExtents : NET::WMGeometry;
+    KWindowInfo winInfo(winId, properties, NET::WM2TransientFor);
 
-    xcb_query_tree_cookie_t treeCookie = xcb_query_tree_unchecked(xcbConn, window);
-    CScopedPointer<xcb_query_tree_reply_t> treeReply(xcb_query_tree_reply(xcbConn, treeCookie, NULL));
-    xcb_window_t *winChildren = xcb_query_tree_children(treeReply.data());
-    int winChildrenLength = xcb_query_tree_children_length(treeReply.data());
-
-    for (int i = winChildrenLength - 1; i >= 0; i--) {
-        winStack.push(winChildren[i]);
+    // add the current window to the image
+    if (mCaptureDecorations) {
+        outRect = winInfo.frameGeometry();
+    } else {
+        outRect = winInfo.geometry();
     }
-
-    return winStack;
+    return winInfo.transientFor();
 }
 
-xcb_window_t X11ImageGrabber::findParent(xcb_window_t window)
+QPoint X11ImageGrabber::getNativeCursorPosition()
 {
+    // QCursor::pos() is not used because it requires additional calculations.
+    // Its value is the offset to the origin of the current screen is in
+    // device-independent pixels while the origin itself uses native pixels.
+
     xcb_connection_t *xcbConn = QX11Info::connection();
+    xcb_query_pointer_cookie_t pointerCookie = xcb_query_pointer_unchecked(xcbConn, QX11Info::appRootWindow());
+    CScopedPointer<xcb_query_pointer_reply_t> pointerReply(xcb_query_pointer_reply(xcbConn, pointerCookie, NULL));
 
-    xcb_query_tree_cookie_t treeCookie = xcb_query_tree_unchecked(xcbConn, window);
-    CScopedPointer<xcb_query_tree_reply_t> treeReply(xcb_query_tree_reply(xcbConn, treeCookie, NULL));
-
-    return treeReply->parent;
+    return QPoint(pointerReply->root_x, pointerReply->root_y);
 }
